/* CG/solution/parts/cgp02.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * This software is partitioned into 16 tasks.                  *
 * Each task can be implemented by one group of the course      *
 * Files: CG/solution/cg_all.c - Solution                       *
 *        CG/skel/cg_01.c      - Skeletons that should be used  *
 *             .. cg_15.c        to implement each task         *
 *        CG/solution/cgp00.c  - Source part of each task       *
 *             ..     cgp16.c                                   *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Felix Triebel                                       *
 *          Michael Speck                                       *
 *          Felix Triebel                                       *
 *          Traugott Streicher                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 02 <<<<<<<<<<<<<<<<<<<<<<*/

/********************** vector routines ************************/

/*------------------------------
 * vector_chunk alloc_vec
 *  IN int needs_halo : flag to indicate the need for extra halo memory
 *  RETURNS : address of the vector data array
 *
 * TASK: allocates the part of a vector-data's memory for the calling process
 *       and initializes it with zero.
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start          : begin of the local chunk
 *  IN int end1           : end+1 of the local chunk
 *  IN int halo_info.size : size of halo data elements
 * USED FUNCTIONS: APPL: save_malloc
 */
vector_chunk alloc_vec(int needs_halo)
{
        int size, I_loc;
        vector_chunk new;
        size = (end1-start) + (needs_halo ? halo_info.size : 0);
        new = (vector_chunk) save_malloc(sizeof(double) * size);
        /* set the vector data to zero */
        for(I_loc=0; I_loc < size; I_loc++) new[I_loc]=0.0;
        return new;
}

/*------------------------------
 * void Vec_Set_Value
 *  IN vector_chunk v : set vector v
 *  IN int I          : at global index I
 *  IN double value   : to value
 *
 * TASK: set one value within a vector
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start   : begin of the local chunk
 *  IN int my_rank : rank of the local process (only for error handling)
 * USED FUNCTIONS: APPL: Row2Rank
 */
void Vec_Set_Value(vector_chunk v, int I, double value)
{
        int I_loc;
        if (Row2Rank(I) != my_rank) { printf("ERROR: Vectorelement not local!\n"); abrt(); }
        I_loc = I - start;
        v[I_loc] = value;
}


/*------------------------------
 * void dupl_vec
 *  IN vector_chunk v1  : vector 1
 *  OUT vector_chunk v2 : vector 2
 *
 * TASK: do a local vector chunk (without halo) duplication: v2 := v1
 *       easy solution: memcpy (see '> man memcpy')
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: C-lib: memcpy
 */
void dupl_vec(vector_chunk v1, vector_chunk v2)
{
        memcpy(v2, v1, sizeof(double)*(end1-start));
}

/*------------------------------
 * void add_vec_vec
 *  IN vector_chunk v1  : vector 1
 *  IN double lambda    : factor lambda
 *  IN vector_chunk v2  : vector 2
 *  OUT vector_chunk v3 : vector 3
 *
 * TASK: calculate locally:   v3 := v1 + lambda * v2
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: none
 */
void add_vec_vec(vector_chunk v1, double lambda, vector_chunk v2,
                 vector_chunk /* out */ v3)
{
        int I_loc;

        for(I_loc=0; I_loc < (end1-start); I_loc++)
                v3[I_loc] = v1[I_loc] + lambda * v2[I_loc];
}

/*------------------------------
 * double dot_product_vec_vec
 *  IN vector_chunk v1 : vector 1
 *  IN vector_chunk v2 : vector 2
 *  RETURNS double     : dot product
 *
 * TASK: calculate global dot product for vector 1 and vector 2
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: none
 *                 MPI : MPI_Allreduce
 */
double dot_product_vec_vec(vector_chunk v1, vector_chunk v2)
{
        int I_loc;
        double sum=0, global_sum;

        /* dot product of local chunk (without halo) */
        for(I_loc=0; I_loc < (end1-start); I_loc++)
                sum += v1[I_loc] * v2[I_loc];

        /* global sum */
#ifdef serial
        global_sum = sum;
#else
        MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#endif

        return global_sum;
}


/*------------------------------
 * double sqr_norm_vec
 *  IN vector_chunk v1 : vector 1
 *  RETURNS double     : square of norm
 *
 * TASK: calculate global (vector norm Euklid)^2 for vector 1
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: none
 *                 MPI : MPI_Allreduce
 */
double sqr_norm_vec(vector_chunk v1)
{
        int I_loc;
        double sum=0, global_sum;

        /* dot product of local chunk (without halo) */
        for(I_loc=0; I_loc < (end1-start); I_loc++)
                sum += v1[I_loc] * v1[I_loc];

        /* global sum */
#ifdef serial
        global_sum = sum;
#else
        MPI_Allreduce(&sum, &global_sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#endif

        return global_sum;
}

/*------------------------------
 * double max_norm_vec
 *  IN vector_chunk v1 : vector 1
 *  RETURNS double     : maximum norm
 *
 * TASK: calculate global (vector maximum norm) for vector 1
 *
 * REFERENCE: Slides 16 (see "own data"), 21
 *
 * GLOBAL DATA:
 *  IN int start : begin of the local chunk
 *  IN int end1  : end+1 of the local chunk
 * USED FUNCTIONS: APPL: none
 *                 MPI : MPI_Allreduce
 */
double max_norm_vec(vector_chunk v1)
{
        int I_loc;
        double maxabs=0, global_max;

        /* dot product of local chunk (without halo) */
        for(I_loc=0; I_loc < (end1-start); I_loc++)
        {
                if ( v1[I_loc] > maxabs) maxabs =  v1[I_loc];
                if (-v1[I_loc] > maxabs) maxabs = -v1[I_loc];
        }

        /* global maximum */
#ifdef serial
        global_max = maxabs;
#else
        MPI_Allreduce(&maxabs, &global_max, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
#endif

        return global_max;
}

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 02 <<<<<<<<<<<<<<<<<<<<<<*/
