/* CG/solution/parts/cgp08.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * This software is partitioned into 16 tasks.                  *
 * Each task can be implemented by one group of the course      *
 * Files: CG/solution/cg_all.c - Solution                       *
 *        CG/skel/cg_01.c      - Skeletons that should be used  *
 *             .. cg_15.c        to implement each task         *
 *        CG/solution/cgp00.c  - Source part of each task       *
 *             ..     cgp16.c                                   *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Felix Triebel                                       *
 *          Michael Speck                                       *
 *          Felix Triebel                                       *
 *          Traugott Streicher                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 08 <<<<<<<<<<<<<<<<<<<<<<*/

#ifndef serial
/*------------------------------
 * void comm_vec_halo
 *
 * IN OUT vector_chunk v1
 *
 * TASK: Communicate halo of a vector: send requested elemets
 *       to others processors and update local halo.
 *
 * REFERENCE: Slides 18, 19, 20, 23
 *
 * USED GLOBAL DATA:
 *   IN int recv_num:                      number of processors we receive data from
 *   IN int* recv_ranks:                   ranks of all real neighbors that are sending halo-data
 *   IN (OUT) neighbor_struct* recv_halo:  (recv_halo[all].scratch modified)
 *   IN int send_num:                      number of processors we send data to
 *   IN int* send_ranks:                   ranks of all real neighbors to which halo-data must be sent
 *   IN (OUT) neighbor_struct* send_halo:  (send_halo[all].scratch modified)
 * USED FUNCTIONS: APPL: save_malloc
 *                 MPI : MPI_Irecv, MPI_ISend, MPI_Waitall
 */
void comm_vec_halo(vector_chunk v1)
{
        int i, j, pos, data_size, src_rank, dest_rank;
        MPI_Request *req;
        MPI_Status *statuses;

        /* allocate scratch memory for req and statuses */ 
        req = (MPI_Request*)save_malloc(sizeof(MPI_Request)*(recv_num+send_num));
        statuses= (MPI_Status*)save_malloc(sizeof(MPI_Status)*(recv_num+send_num));

        /* receive values from neighbors */
        for (i=0; i<recv_num; i++)
        {
                src_rank = recv_ranks[i];
                data_size = recv_halo[src_rank].size;
                /* receive data non-blocking */
                MPI_Irecv(recv_halo[src_rank].scratch, data_size,
                         MPI_DOUBLE, src_rank, MSG_TAG2,
                         MPI_COMM_WORLD, &(req[i]));
        }

        /* send values to neighbors */
        for (i=0; i<send_num; i++)
        {
                dest_rank = send_ranks[i];
                data_size = send_halo[dest_rank].size;
                /* fill send_halo[for all].scratch[] */
                for (j=0; j<data_size; j++)
                {
                        pos = send_halo[dest_rank].I_loc[j];
                        send_halo[dest_rank].scratch[j] = v1[pos];
                }
                /* send data non-blocking */
                MPI_Isend(send_halo[dest_rank].scratch, data_size,
                           MPI_DOUBLE, dest_rank, MSG_TAG2,
                           MPI_COMM_WORLD, &(req[recv_num+i]));
                                            //   ^__ use second part of req array
        }

        /* Waitall */
        if (recv_num+send_num > 0)
                MPI_Waitall(recv_num+send_num, req, statuses);

        /* receive values from neighbors */
        for (i=0; i<recv_num; i++)
        {
                src_rank = recv_ranks[i];
                data_size = recv_halo[src_rank].size;
                for (j=0; j<data_size; j++)
                {
                        pos = recv_halo[src_rank].I_loc[j];
                        v1[pos] = recv_halo[src_rank].scratch[j];
                }
        }
 
        /* free scratch memory */ 
        free(req); free(statuses);
}
#endif

/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 08 <<<<<<<<<<<<<<<<<<<<<<*/
