/* CG/solution/parts/cgp10.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * This software is partitioned into 16 tasks.                  *
 * Each task can be implemented by one group of the course      *
 * Files: CG/solution/cg_all.c - Solution                       *
 *        CG/skel/cg_01.c      - Skeletons that should be used  *
 *             .. cg_15.c        to implement each task         *
 *        CG/solution/cgp00.c  - Source part of each task       *
 *             ..     cgp16.c                                   *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Felix Triebel                                       *
 *          Michael Speck                                       *
 *          Felix Triebel                                       *
 *          Traugott Streicher                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 10 <<<<<<<<<<<<<<<<<<<<<<*/

/******************* domain decomposition **********************/

/*----------------------- 1-dim decomposition -----------------*/

//   Example: m=5, n=7
//
//       |       j=0     1     2     3     4     5     6 |
//   ----+-----------------------------------------------+
//       |      I=                                       |
//   i=0 |         0     1     2     3     4     5     6 |
//     1 |         7     8     9    10    11    12    13 |
//     2 |        14    15    16    17    18    19    20 |
//     3 |        21    22    23    24    25    26    27 |
//     4 |        28    29    30    31    32    33    34 |
//   ----+-----------------------------------------------+

/*------------------------------
 * Init_1dim
 *
 * TASK: None (except printing).
 *
 * REFERENCE: Slides 6, 26
 *
 * USED GLOBAL DATA: none
 * USED FUNCTIONS: none
 */
void Init_1dim(void)
{
        if ((print_level >= 1) && (my_rank == 0))
            printf("Using 1-dim domain decomposition, numprocs=%d\n", numprocs);
}


/*------------------------------
 * I2ij_1dim
 *      IN  I:  global matrix row index
 *      OUT i:  vertical   index of the physical Laplace problem
 *      OUT j:  horizontal index of the physical Laplace problem
 *
 * TASK: Calculates i and j according the scheme in the example above.
 *
 * REFERENCE: Slides 6, 26
 *
 * USED GLOBAL DATA: 
 *  IN int n: vertical size of the physical Laplace problem
 * USED FUNCTIONS: none
 */
void I2ij_1dim(int I, int* i, int* j)
{
        (*i) = I/n; (*j) = I - (*i)*n;
}

/*------------------------------
 * ij2I_1dim
 *      IN  i:  vertical   index of the physical Laplace problem
 *      IN  j:  horizontal index of the physical Laplace problem
 *      OUT I:  global matrix row index
 *
 * TASK: Calculates I according the scheme in the example above.
 *
 * REFERENCE: Slides 6, 26
 *
 * USED GLOBAL DATA: 
 *  IN int n: vertical size of the physical Laplace problem
 * USED FUNCTIONS: none
 */
void ij2I_1dim(int i, int j, int* I)
{
        (*I) = i*n + j;
}

/*----------------------- 2-dim decomposition -----------------*/

//   Example: m=5, n=7, m_procs=2, n_procs=2
//
//       |       j=0     1     2     3  |  4     5     6 |
//   ----+------------------------------+----------------+  v
//       |      I=       (jj=0)               (jj=1)     |
//   i=0 |         0     1     2     3  | 12    13    14 |  i_
//       |                                               |  chunk_
//     1 | (ii=0)  4     5     6     7  | 15    16    17 |  size
//       |                                               |  = 3
//     2 |         8     9    10    11  | 18    19    20 |
//   - - + - -  - - - - - - - - - - - - + - - - - - - - -+  ^
//     3 |        21    22    23    24  | 29    30    31 |
//       | (ii=1)                                        |
//     4 |        25    26    27    28  | 32    33    34 |
//   ----+------------------------------+----------------+
//
//       >    j_chunk_size = 4     <
//
// Caution: Only if the m and n are multiples of _proc and m_proc
//          then the default distibution fits to the domain
//          decomposition. In this example it does not fit:
//          Process 0 works on I =  0.. 8
//          Process 1 works on I =  9..17
//          Process 2 works on I = 18..26
//          Process 3 works on I = 27..34

int i_chunk_size, j_chunk_size, i_n_size;

/*------------------------------
 * Init_2dim
 *
 * TASK: None (except printing).
 *
 * REFERENCE: Slides 26, 24
 *
 * USED GLOBAL DATA: 
 *  OUT int i_chunk_size, j_chunk_size: see figure above
 *  OUT int i_n_size := i_chunk_size * n
 * USED FUNCTIONS: none
 */
void Init_2dim(void)
{
        if ((print_level >= 1) && (my_rank == 0))
            printf("Using 2-dim domain decomposition, m_procs=%d, n_procs=%d, numprocs=%d\n",
                    m_procs, n_procs, numprocs);
        i_chunk_size = (m-1)/m_procs + 1; /* rounded up*/
        i_n_size = i_chunk_size * n;
        j_chunk_size = (n-1)/n_procs + 1; /* rounded up*/
}


/*------------------------------
 * I2ij_2dim
 *      IN  I:  global matrix row index
 *      OUT i:  vertical   index of the physical Laplace problem
 *      OUT j:  horizontal index of the physical Laplace problem
 *
 * TASK: Calculates i and j according the scheme in the example above.
 *
 * REFERENCE: Slides 26, 24
 *
 * USED GLOBAL DATA: 
 *  IN int i_chunk_size, j_chunk_size: see figure above
 *  IN int i_n_size := i_chunk_size * n
 *  IN int m: horizontal size of the physical Laplace problem
 *  IN int n: vertical   size of the physical Laplace problem
 * USED FUNCTIONS: none
 */
void I2ij_2dim(int I, int* i_arg, int* j_arg)
{
        int i, j, ii, jj, m_slice, n_slice, i_j_size, I_tail;

        i = 0;
        ii = I / i_n_size;
        i = i + ii * i_chunk_size;
        m_slice = m - ii * i_chunk_size;
        if (m_slice > i_chunk_size) m_slice = i_chunk_size;
        I_tail = I - ii * i_n_size;

        j = 0;
        i_j_size = m_slice * j_chunk_size;
        jj = I_tail / i_j_size;
        j = j + jj * j_chunk_size;
        I_tail = I_tail - jj * i_j_size;
        n_slice = n - jj * j_chunk_size;
        if (n_slice > j_chunk_size) n_slice = j_chunk_size;

        i = i + I_tail / n_slice;
        j = j + I_tail % n_slice;

        (*i_arg) = i;   (*j_arg) = j;
}

/*------------------------------
 * ij2I_2dim
 *      IN  i:  vertical   index of the physical Laplace problem
 *      IN  j:  horizontal index of the physical Laplace problem
 *      OUT I:  global matrix row index
 *
 * TASK: Calculates I according the scheme in the example above.
 *
 * REFERENCE: Slides 26, 24
 *
 * USED GLOBAL DATA: 
 *  IN int i_chunk_size, j_chunk_size: see figure above
 *  IN int i_n_size := i_chunk_size * n
 *  IN int m: horizontal size of the physical Laplace problem
 *  IN int n: vertical   size of the physical Laplace problem
 * USED FUNCTIONS: none
 */
void ij2I_2dim(int i, int j, int* I)
{
        int ii, jj, i_tail, j_tail, m_slice, n_slice;

        ii     = i / i_chunk_size;    jj     = j / j_chunk_size;
        i_tail = i % i_chunk_size;    j_tail = j % j_chunk_size;
        m_slice = m - ii * i_chunk_size;
        if (m_slice > i_chunk_size) m_slice = i_chunk_size;
        n_slice = n - jj * j_chunk_size;
        if (n_slice > j_chunk_size) n_slice = j_chunk_size;
        (*I) = ii*i_n_size + jj*(m_slice * j_chunk_size) + i_tail*n_slice + j_tail;
}

/*-------- switching between 1-dim and 2-dim decomposition -------*/
void Init_Decomp(void)                    /* description, see Init_1dim() above */
{       switch (decomp_dims)
        {       case 1: Init_1dim(); break;
                case 2: Init_2dim(); break;
                default: printf("Wrong decomp_dims=%d\n",decomp_dims); abrt();
        }
}
void I2ij(int I, int* i_arg, int* j_arg)  /* description, see I2ij_1dim() above */
{       switch (decomp_dims)
        {       case 1: I2ij_1dim(I,i_arg,j_arg); break;
                case 2: I2ij_2dim(I,i_arg,j_arg); break;
                default: printf("Wrong decomp_dims=%d\n",decomp_dims); abrt();
        }
}
void ij2I(int i, int j, int* I)           /* description, see ij2I_1dim() above */
{       switch (decomp_dims)
        {       case 1: ij2I_1dim(i,j,I); break;
                case 2: ij2I_2dim(i,j,I); break;
                default: printf("Wrong decomp_dims=%d\n",decomp_dims); abrt();
        }
}
/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 10 <<<<<<<<<<<<<<<<<<<<<<*/
